package final24;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Scanner;
import java.io.File;
import java.util.regex.*;
import java.io.FileInputStream;
import com.sun.speech.freetts.*;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Random;
import java.util.Stack;

public class Main
{
    public static void main(String[] args) throws FileNotFoundException, IOException
    {
        HashMap<String, Integer> syllables = new HashMap<String, Integer>();
        HashMap<String, String> rhymes = new HashMap<String, String>();

        System.out.println("Reading the rhyming dictionary into syllable counts and rhymes...");
        FileInputStream file = new FileInputStream(new File("dictionary.txt"));
        Scanner read = new Scanner(file);
        while (read.hasNext()) {
            String[] parts = read.nextLine().split("  ");

            String word = parts[0];
            String pronunciation = parts[1];

            syllables.put(word, countSyllables(pronunciation));
            rhymes.put(word, findRhyme(pronunciation));
        }
        file.close();

        // Create our markov chain generator.
        ChainGenerator<String> sentences = new ChainGenerator<String>(2);

        System.out.println("Reading all lines of War and Peace into the markov chain generator...");
        file = new FileInputStream(new File("corpus.txt"));
        read = new Scanner(file);
        while (read.hasNext()) {
            String[] words = read.nextLine().split(" ");
            reverse(words);
            sentences.add(words);
        }
        file.close();

        System.out.println("Building the limerick...");
        Random r = new Random();
        String[] L1, L2, L3, L4, L5;
        int i = 0;
        do {
            L1 = buildLine(sentences, r, syllables, rhymes, 8);
            L2 = buildLine(sentences, r, syllables, rhymes, 8, L1);
            L3 = buildLine(sentences, r, syllables, rhymes, 5);
            L4 = buildLine(sentences, r, syllables, rhymes, 5, L3);
            L5 = buildLine(sentences, r, syllables, rhymes, 8, L1, L2);
            i++;
        } while (L2 == null || L4 == null || L5 == null);  // Try building again, if we failed to create a line.

        // Turn the limerick into strings.
        String LS1 = join(L1, " ");
        String LS2 = join(L2, " ");
        String LS3 = join(L3, " ");
        String LS4 = join(L4, " ");
        String LS5 = join(L5, " ");

        // Allocate a voice.
        System.out.println();
        VoiceManager voiceManager = VoiceManager.getInstance();
        Voice voice = voiceManager.getVoice("kevin16");
        voice.allocate();
        System.out.println();

        // Print and speak the limerick.
        System.out.println(LS1); voice.speak(LS1);
        System.out.println(LS2); voice.speak(LS2);
        System.out.println(LS3); voice.speak(LS3);
        System.out.println(LS4); voice.speak(LS4);
        System.out.println(LS5); voice.speak(LS5);

        // Clean up.
        voice.deallocate();
    }

    private static String[] buildLine(ChainGenerator<String> gen, Random r, HashMap<String, Integer> syll, HashMap<String, String> rhy, int target)
    {
        Stack<String> result = buildPhrase(gen, r, syll, new ChainState<String>(new String[0]), target);
        if (result == null) return null;

        Object[] res = result.toArray();
        String[] str = new String[res.length];
        System.arraycopy(res, 0, str, 0, res.length);
        return str;
    }

    private static Stack<String> buildPhrase(ChainGenerator<String> gen, Random r, HashMap<String, Integer> syll, ChainState<String> state, int depth)
    {
        if (depth == 0)
        {
            Integer terminal = gen.getTerminal(state);
            if (terminal > 0) return new Stack<String>();
        }
        else
        {
            LinkedList<String> nexts = scramble(r, gen.getNextStates(state));

            for (String n : nexts)
            {
                String word = normalize(n);
                if (!syll.containsKey(word)) continue;

                int newDepth = depth - syll.get(word);
                if (newDepth < 0) continue;

                Stack<String> result = buildPhrase(gen, r, syll, state.push(n, gen.order), newDepth);
                if (result == null) continue;

                result.push(n);
                return result;
            }
        }

        return null;
    }

    private static String[] buildLine(ChainGenerator<String> gen, Random r, HashMap<String, Integer> syll, HashMap<String, String> rhy, int target, String[] match, String[] ... exclude)
    {
        String rhyme = rhy.get(normalize(match[match.length - 1]));

        HashSet<String> excl = new HashSet<String>();
        excl.add(normalize(match[match.length - 1]));
        for (String[] ex : exclude) {
            if (ex == null) return null;
            excl.add(normalize(ex[ex.length - 1]));
        }

        LinkedList<String> init = scramble(r, gen.getInitialStates());
        for (String n : init)
        {
            String word = normalize(n);
            if (excl.contains(word)) continue;
            if (!syll.containsKey(word)) continue;
            if (!rhyme.equals(rhy.get(word))) continue;

            int depth = target - syll.get(word);
            if (depth < 0) continue;

            Stack<String> result = buildPhrase(gen, r, syll, new ChainState<String>(new String[] { n }), depth);
            if (result == null) continue;

            result.push(n);
            Object[] res = result.toArray();
            String[] str = new String[res.length];
            System.arraycopy(res, 0, str, 0, res.length);
            return str;
        }

        return null;
    }

    private static void reverse(String[] array) {
        for(int i = 0; i < array.length / 2; i++)
        {
            String temp = array[i];
            array[i] = array[array.length - i - 1];
            array[array.length - i - 1] = temp;
        }
    }

    private static Pattern vowel = Pattern.compile("[A-Z]{2}\\d");

    private static String findRhyme(String pronunciation) {
        String[] phonemes = pronunciation.split(" ");

        int rhymeStart = 0;
        for (int i = phonemes.length - 1; i > 0; i--) {
            if (phonemes[i].matches(vowel.pattern())) { rhymeStart = i; break; }
        }

        StringBuilder rhyme = new StringBuilder();
        for (int i = rhymeStart; i < phonemes.length; i++) {
            if (rhyme.length() > 0) rhyme.append(" ");
            rhyme.append(phonemes[i]);
        }

        return rhyme.toString();
    }

    private static int countSyllables(String pronunciation) {
        Matcher vowels = vowel.matcher(pronunciation);

        int syllables = 0;
        while (vowels.find()) syllables++;
        return syllables;
    }

    public static String join(String[] s, String delimiter) {
        StringBuilder buffer = new StringBuilder();
        for (String i : s) {
            if (buffer.length() > 0) buffer.append(delimiter);
            buffer.append(i);
        }
        return buffer.toString();
    }

    private static LinkedList<String> scramble(Random r, HashMap<String, Integer> weights) {
        LinkedList<String> items = new LinkedList<String>();

        if (weights == null) return items;

        int total = 0;
        for (int w : weights.values())
        {
            total += w;
        }

        while (!weights.isEmpty())
        {
            int value = r.nextInt(total) + 1;

            int currentWeight = 0;
            for (String itemKey : weights.keySet())
            {
                int itemValue = weights.get(itemKey);
                currentWeight += itemValue;
                if (currentWeight >= value)
                {
                    items.add(itemKey);
                    total -= itemValue;
                    weights.remove(itemKey);
                    break;
                }
            }
        }

        return items;
    }

    private static String normalize(String n) {
        return n.toUpperCase().replaceAll("[^A-Z]+$", "");
    }
}
